using System;
using System.IO;
using System.Text.RegularExpressions;
using Antlr.StringTemplate;

namespace InSiteTransactionProfiler
{
	class txnprof
	{
		static TimeSpan utc_span = DateTime.Now-DateTime.UtcNow;
		static InSiteTxnLog g_CDOs = new InSiteTxnLog();

		static RegexOptions matchOpt = RegexOptions.Compiled | RegexOptions.Singleline;

		// match eg: "Depth: 1 CDO TestLotSchedule Event: AfterInitialize"
		static Regex reCDOEvent = new Regex(@"Depth: (?<depth>\w+) CDO:? (?<cdo>\w+) Event: (?<event>\w+)",matchOpt);
		// match eg: "Depth: 3 CDO: TestLotSchedule Perform: PreValidate"
		static Regex reCDOPerform = new Regex(@"Depth: (?<depth>\w+) CDO: (?<cdo>\w+) Perform: (?<perform>\w+)",matchOpt);
		// match eg: "Depth: 1 CDO: TestLotSchedule Field: Constants Event: Initialize"
		static Regex reCDOFieldEvent = new Regex(@"Depth: (?<depth>\w+) CDO: (?<cdo>\w+) Field: (?<field>\w+) Event: (?<event>\w+)",matchOpt);

		// match eg: "Exec: Camstar.Call.1"
		static Regex reExec =  new Regex(@"Exec: Camstar.(?<exec>\w+).1",matchOpt);
		// match eg: "Name: Event Value: PreValidate"
		static Regex reParameter = new Regex(@"Name: (?<name>\w+) Value: (?<value>.+)",matchOpt);

		[STAThread]
		static int Main(string[] args)
		{		
			if (args.Length != 1)
				return PrintUsage();

			#region Read transaction log file into memory
			using (StreamReader sr = new StreamReader(args[0],System.Text.Encoding.Unicode))
			{
				string t;

				// Handle transaction log header
				t = Trim_DateTimeAndProcessID(sr.ReadLine(),out g_CDOs.service_start_time,out g_CDOs.service_start_pcs_id);
				g_CDOs.service_name = t.Substring(21, t.IndexOf(' ',21)-21);
				g_CDOs.user_name = t.Substring(t.IndexOf(' ',21)+4);

				DateTime startTime;
				int pcs_id;

				t = Trim_DateTimeAndProcessID(sr.ReadLine(), out startTime, out pcs_id);
				CDOInfo cdoInfo = null;
				bool cdoInfoAdded = false;
				while(t != string.Empty)
				{
					Match matchResult;


					matchResult = reCDOEvent.Match(t);
					if(matchResult.Success == true)
					{
						string depth, cdo, str_event;
						depth = matchResult.Result("${depth}");
						cdo = matchResult.Result("${cdo}");
						str_event = matchResult.Result("${event}");

						cdoInfo = new CDOEvent(str_event,startTime, pcs_id, int.Parse(depth),cdo);
						cdoInfoAdded = false;
					}

					matchResult = reCDOPerform.Match(t);
					if(matchResult.Success == true)
					{
						string depth, cdo, perform;
						depth = matchResult.Result("${depth}");
						cdo = matchResult.Result("${cdo}");
						perform = matchResult.Result("${perform}");

						cdoInfo = new CDOPerform(perform, startTime, pcs_id, int.Parse(depth),cdo);
						cdoInfoAdded = false;
					}

					matchResult = reCDOFieldEvent.Match(t);
					if(matchResult.Success == true)
					{
						string depth, cdo, field, str_event;
						depth = matchResult.Result("${depth}");
						cdo = matchResult.Result("${cdo}");
						field = matchResult.Result("${field}");
						str_event = matchResult.Result("${event}");

						cdoInfo = new CDOFieldEvent(str_event,field,startTime,pcs_id,int.Parse(depth),cdo);
						cdoInfoAdded = false;
					}

					bool execAdded = false;
					do
					{
						if(reExec.Match(t).Success == false && execAdded == false)
						{
							t = Trim_DateTimeAndProcessID(sr.ReadLine(), out startTime, out pcs_id);
						}
						matchResult = reExec.Match(t);
						if(matchResult.Success == true)
						{
							Exec newExec = new Exec(matchResult.Result("${exec}"),startTime,pcs_id);
							// Fill parameter
							do
							{
								t = Trim_DateTimeAndProcessID(sr.ReadLine(), out startTime, out pcs_id);
								matchResult = reParameter.Match(t);
								if(matchResult.Success == true)
								{									
									newExec.paramName.Add(matchResult.Result("${name}"));
									newExec.paramValue.Add(matchResult.Result("${value}"));
								}
								else
									break;
							}while(true);								
							cdoInfo.execs.Add(newExec);
							execAdded=true;
						}
						else
							break;
					}while(true);
					if(cdoInfoAdded == false)
					{
						g_CDOs.Add(cdoInfo);
						cdoInfoAdded = true;
					}
					if(t.StartsWith("deleting transaction") == true)
					{
						g_CDOs.service_end_time = startTime;
						g_CDOs.service_end_pcs_id = pcs_id;
					}
				}
			}
			#endregion


			StringTemplateGroup group = new StringTemplateGroup("txnprof");
			StringTemplate st = group.GetInstanceOf("txnprof");

			// Fill Execution Summary
			st.SetAttribute("TITLE", string.Format("txnprof Output of the InSite Transaction Log file {0}", args[0]));
			st.SetAttribute("TOTAL_CDO_CALLED", g_CDOs.Count);
			st.SetAttribute("SERVICE_NAME", g_CDOs.service_name);
			st.SetAttribute("USER_NAME", g_CDOs.user_name);
			st.SetAttribute("SERVICESTARTPCSID", g_CDOs.service_start_pcs_id);
			st.SetAttribute("SERVICESTARTTIME", milsecDateTime(g_CDOs.service_start_time));
			st.SetAttribute("LASTCDOENDTIME", milsecDateTime(g_CDOs.LastCDOEndTime));
			st.SetAttribute("LASTCDOTIMESPAN", g_CDOs.LastCDOEndTime - g_CDOs.service_start_time);
			st.SetAttribute("SERVICEENDPCSID", g_CDOs.service_end_pcs_id);
			st.SetAttribute("SERVICEENDTIME", milsecDateTime(g_CDOs.service_end_time));
			st.SetAttribute("TOTALSERVICETIMESPAN", g_CDOs.service_end_time - g_CDOs.service_start_time);			

			// Fill Called CDO
			for(int i=0; i<g_CDOs.Count; i++)
			{
				DateTime endTime;
				if((i+1)==g_CDOs.Count)
					endTime = g_CDOs.service_end_time;
				else
					endTime = g_CDOs[i+1].startTime;
				st.SetAttribute("GeneralCDOInfo", GeneralCDOInfoInTable(i+1, g_CDOs[i],endTime));
//			}
//
			// Fill transaction log
//			for(int i=0; i<g_CDOs.Count; i++)
//			{
				int depth = g_CDOs[i].depth;
				st.SetAttribute("DETAILLOGS", GeneralCDOInfoWithDepth(depth,g_CDOs[i],i+1));
				foreach(Exec e in g_CDOs[i].execs)
				{
					st.SetAttribute("DETAILLOGS", GeneralExecInfo(depth+1, e));					
				}
			}

			// Writing the HTML output
			using (StreamWriter sw = new StreamWriter("txnprof.html",false, System.Text.Encoding.Unicode))
			{
				sw.Write(st.ToString());
			}

			System.Diagnostics.Process.Start("txnprof.html");
			return 0;
		}

		static string GeneralCDOInfoWithDepth(int depth, CDOInfo cdo,int CDONo)
		{
			string depStr = new string(' ',depth*4);
			if (cdo is CDOEvent)
			{
				CDOEvent c = (CDOEvent) cdo;
				return string.Format("<a name={5}>{0}Event: {3} CDOType:{1} CDOName:{2} Depth:{4}\r\n</a>",depStr, c.GetType().Name, c.CDOname, c.eventName, c.depth,CDONo);
			}
			else if (cdo is CDOFieldEvent)
			{
				CDOFieldEvent c = (CDOFieldEvent) cdo;
				return string.Format("<a name={6}>{0}Field: {3} Event:{4} CDOType:{1} CDOName:{2} Depth:{5}\r\n</a>",depStr, c.GetType().Name, c.CDOname, c.fieldName, c.eventName, c.depth,CDONo);
			}
			else if (cdo is CDOPerform)
			{
				CDOPerform c = (CDOPerform) cdo;
				return string.Format("<a name={5}>{0}Perform: {3} CDOType:{1} CDOName:{2} Depth:{4}\r\n</a>",depStr, c.GetType().Name, c.CDOname, c.performName, c.depth,CDONo);
			}
			else
			{
				return string.Format("Unknow CDOInfo type {0}\r\n",cdo.GetType().Name);
			}
		}

		static string GeneralExecInfo(int depth, Exec e)
		{
			System.Text.StringBuilder sb = new System.Text.StringBuilder(100);
			sb.AppendFormat("{0}{1}(",new System.String(' ',depth*4), e.execName);
			for(int i=0; i<e.paramName.Count; i++)
			{
				if(e.paramName[i] != "__ExeIndex" && e.paramName[i] != "__FunctionId")
				{
					sb.AppendFormat("{0}=>{1}, ", e.paramName[i], e.paramValue[i]);
				}
			}
			if(e.paramName.Count > 2)
				sb.Remove(sb.Length-2,2);
			sb.Append(")\r\n");
			return sb.ToString();
		}

		static string GeneralCDOInfoInTable(int seq, CDOInfo cdo, DateTime endTime)
		{
			TimeSpan et = endTime - cdo.startTime;			
			if(cdo is CDOEvent)
			{
				CDOEvent c = (CDOEvent) cdo;
				return string.Format("<TR><TD>{5}</TD><TD>{0}</TD><TD>{1}</TD><TD>&nbsp;</TD><TD><a href=\"#{5}\">{2}</a></TD><TD>&nbsp;</TD><TD>{3}</TD><TD>{4}</TD></TR>\r\n",c.depth,c.CDOname,c.eventName,c.execs.Count, et.TotalMilliseconds,seq);
			}
			else if(cdo is CDOFieldEvent)
			{
				CDOFieldEvent c = (CDOFieldEvent) cdo;
				return string.Format("<TR><TD>{6}</TD><TD>{0}</TD><TD>{1}</TD><TD><a href=\"#{6}\">{2}</a></TD><TD>{3}</TD><TD>&nbsp;</TD><TD>{4}</TD><TD>{5}</TD></TR>\r\n",c.depth,c.CDOname,c.fieldName,c.eventName,c.execs.Count,et.TotalMilliseconds, seq);
			}
			else if(cdo is CDOPerform)
			{
				CDOPerform c = (CDOPerform) cdo;
				return string.Format("<TR><TD>{5}</TD><TD>{0}</TD><TD>{1}</TD><TD>&nbsp;</TD><TD>&nbsp;</TD><TD><a href=\"#{5}\">{2}</a></TD><TD>{3}</TD><TD>{4}</TD></TR>\r\n",c.depth,c.CDOname,c.performName,c.execs.Count,et.TotalMilliseconds, seq);
			}
			else
			{
				return string.Format("Unknow CDOInfo type {0}",cdo.GetType().Name);
			}
		}

		static string Trim_DateTimeAndProcessID(string line_in, out DateTime dt, out int pcs_id)
		{
			if( line_in == null )
			{
					dt = DateTime.MinValue;
				pcs_id = int.MinValue;
				return string.Empty;
			}
			dt = To_DateTime(line_in.Substring(0,line_in.IndexOf('\t')));
			pcs_id = int.Parse(line_in.Substring(19, line_in.IndexOf('\t',19)-19));
			return line_in.Substring(24);
		}

		static DateTime To_DateTime(string dt)
		{ 
			int year = int.Parse(dt.Substring(0,4));
			int month = int.Parse(dt.Substring(4,2));
			int day = int.Parse(dt.Substring(6,2));
			int hour = int.Parse(dt.Substring(8,2));
			int min = int.Parse(dt.Substring(10,2));
			int sec = int.Parse(dt.Substring(12,2));
			int millsec = int.Parse(dt.Substring(15,3));
			return new DateTime(year,month,day,hour,min,sec,millsec)+utc_span;
		}

		static string milsecDateTime(DateTime dt)
		{
			return string.Format("{0}.{1}",dt.ToString("u"),dt.Millisecond);
		}

		static int PrintUsage()
		{
			Console.WriteLine("txnprof [insite_transaction.log]"); 
			return 1; 
		}
	}
}
